#include <nds.h>
#ifndef _NOTDS
#include <dswifi7.h>
#include <maxmod7.h>
#include <stdio.h>
#include <stdbool.h>

#define f32 int

#define SOUND_FREQ(n)	((-0x1000000 / (n)))

typedef struct {
	bool stereo;
	uint sampleRate;
	int bytesPerSample;
	uint dataSize;
	bool streamed;
	int fOffset;
	bool ogg;
	union {
		unsigned char* samples;
		short* shortSamples;
	};
	FILE* streamFile;
} SoundEffect;

typedef struct {
	f32 pan;
	f32 volume;
	f32 pitch;
	bool loop;
	unsigned int loopEnd;
	unsigned int loopStart;
	SoundEffect* sound;
} SoundData;

typedef struct {
	SoundData* data;
	volatile bool done;
	volatile int id;
} FIFOAudio;

typedef struct {
	volatile unsigned int SOUNDCNT;
	volatile unsigned int SOUNDSAD;
	volatile unsigned short SOUNDTMR;
	volatile unsigned short SOUNDPNT;
	volatile unsigned int SOUNDLEN;
} AudioRegister;

//---------------------------------------------------------------------------------
void VblankHandler(void) {
//---------------------------------------------------------------------------------
	Wifi_Update();
}


//---------------------------------------------------------------------------------
void VcountHandler() {
//---------------------------------------------------------------------------------
	inputGetAndSend();
}

volatile bool exitflag = false;

//---------------------------------------------------------------------------------
void powerButtonCB() {
//---------------------------------------------------------------------------------
	exitflag = true;
}

void PlayAudioCallback(void* address, void* userdata) {
	FIFOAudio* data = (FIFOAudio*)address;
	AudioRegister* audioRegisters = 0x04000400;
	int currRegister = 0;
	// get free audio register
	for (int i = 0; i < 16; ++i) {
		if (audioRegisters[i].SOUNDCNT & 0x80000000) {
			// failed to find free register...
			if (i == 15) {
				currRegister = -1;
				break;
			}
			continue;
		}
		currRegister = i;
		break;
	}

	// bail out early, no register was found
	if (currRegister == -1) {
		data->done = true;
		fifoSendValue32(FIFO_USER_01, -1);
		return;
	}

	// register was found, employ data
	audioRegisters[currRegister].SOUNDSAD = data->data->sound->samples;
	// mark bit to start playing
	unsigned int workValue = 0x80000000;

	// volume
	workValue |= data->data->volume / 33;

	// panning
	workValue |= (data->data->pan / 33) << 16;
	// loop
	if (data->data->loop) {
		workValue |= 1 << 27;
	}
	else {
		workValue |= 2 << 27;
	}
	// format
	workValue |= (data->data->sound->bytesPerSample - 1) << 29;

	// loop point
	audioRegisters[currRegister].SOUNDPNT = data->data->loopStart;
	// audio length
	audioRegisters[currRegister].SOUNDLEN = data->data->sound->dataSize / 4;

	// frequency maybe?
	audioRegisters[currRegister].SOUNDTMR = SOUND_FREQ((int)data->data->sound->sampleRate);

	// aaand go!
	audioRegisters[currRegister].SOUNDCNT = workValue;

	data->id = currRegister;
	fifoSendValue32(FIFO_USER_01, currRegister);
	data->done = true;
}

//---------------------------------------------------------------------------------
int main() {
//---------------------------------------------------------------------------------
	// clear sound registers
	dmaFillWords(0, (void*)0x04000400, 0x100);

	REG_SOUNDCNT |= SOUND_ENABLE;
	writePowerManagement(PM_CONTROL_REG, ( readPowerManagement(PM_CONTROL_REG) & ~PM_SOUND_MUTE ) | PM_SOUND_AMP );
	powerOn(POWER_SOUND);

	readUserSettings();
	ledBlink(0);

	irqInit();
	// Start the RTC tracking IRQ
	initClockIRQ();
	fifoInit();
	touchInit();

	mmInstall(FIFO_MAXMOD);

	SetYtrigger(80);

	installWifiFIFO();
	installSoundFIFO();

	installSystemFIFO();

	irqSet(IRQ_VCOUNT, VcountHandler);
	irqSet(IRQ_VBLANK, VblankHandler);

	irqEnable( IRQ_VBLANK | IRQ_VCOUNT | IRQ_NETWORK);

	setPowerButtonCB(powerButtonCB);

	fifoSetAddressHandler(FIFO_USER_01, PlayAudioCallback, NULL);

	// Keep the ARM7 mostly idle
	while (!exitflag) {
		swiWaitForVBlank();
	}
	return 0;
}
#endif