#include "sdsound.h"
#include "sdfile.h"
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#ifdef _WIN32
#include "PC/pcaudio.h"
#endif
int currSoundId;
SoundEffect* currMusic;
unsigned char* mainMusic;
int currMusicId;
#ifndef _NOTDS
#include <nds/ndstypes.h>
#include "sdmath.h"

bool playingMusic;
PlayingSoundData currMusicData;

int sfxIds[32];

int streamId;


typedef struct {
	SoundData* data;
	volatile bool done;
	volatile int id;
} FIFOAudio;

FIFOAudio FIFOData;

int prevTimerRemainder;
int prevTimerPos;
int prevMusicRemainder;
int prevSampleRateRemainder;

int musicId;

void MusicCallback(int length, char* dest) {
	int readSize = 1;
	if (currMusic->stereo) {
		readSize = 2;
	}
	readSize *= currMusic->bytesPerSample;
	if (!currMusic->ogg) {
		if (currMusicData.soundPosition + length * readSize >= (currMusicData.loopEnd * readSize)) {
			int readDeficit = (currMusicData.loopEnd * readSize) - currMusicData.soundPosition;
			fread(dest, readDeficit, readSize, currMusic->streamFile);
			fseek(currMusic->streamFile, (currMusicData.loopStart * readSize) + currMusic->fOffset, SEEK_SET);
			fread(dest + readDeficit, length - readDeficit, readSize, currMusic->streamFile);
		}
		else {
			fread(dest, length, readSize, currMusic->streamFile);
		}
		if (currMusic->bytesPerSample == 1) {
			char* charDest = (char*)dest;
			const int len = length * readSize;
			for (int i = 0; i < len; ++i) {
				charDest[i] += 128;
			}
		}
		currMusicData.soundPosition = ftell(currMusic->streamFile) - currMusic->fOffset;
	}
}

void MusicTimerCallback() {
	int elapsed = 1024;
	int startRead = prevTimerPos;
	int endRead = prevTimerPos + elapsed;
	int lastSample = 4096;
		if (endRead > lastSample) {
		endRead = endRead - lastSample;
	}
	// do two reads if we're overflowing
	if (endRead < startRead) {
		MusicCallback(lastSample - startRead, (char*)&currMusic->samples[startRead]);
		MusicCallback(endRead, (char*)currMusic->samples);
	}
	else {
		MusicCallback(endRead - startRead, (char*)&currMusic->samples[startRead]);
	}
	prevTimerPos = endRead;
}

void PlayMusic(char* filedir) {
	if (currMusic != NULL) {
		StopMusic();
	}
	PlayingSoundData sd;
	sd.pan = 2048;
	sd.pitch = 4096;
	sd.volume = 4096;
	sd.sound = currMusic;
	sd.loop = true;
	sd.soundPosition = 0;
	if (strcmp(".wav", &filedir[strlen(filedir) - 4]) == 0) {
		currMusic = LoadWavStreamed(filedir);
	}
	if (currMusic == NULL) {
		return;
	}
	sd.sound = currMusic;
	char loopFileDir[512];
	sprintf(loopFileDir, "%s%s", filedir, ".loop");
	FILE* loopFile = fopen(loopFileDir, "rb");
	if (loopFile != NULL) {
		fread(&sd.loopStart, 4, 1, loopFile);
		fread(&sd.loopEnd, 4, 1, loopFile);
		fclose(loopFile);
	}
	else {
		sd.loopStart = 0;
		fseek(currMusic->streamFile, 0, SEEK_END);
		sd.loopEnd = ftell(currMusic->streamFile) - currMusic->fOffset;

		int readSize = 1;
		if (currMusic->stereo) {
			readSize = 2;
		}
		readSize *= currMusic->bytesPerSample;
		sd.loopEnd /= readSize;

		fseek(currMusic->streamFile, 0, SEEK_SET);
	}
	fseek(currMusic->streamFile, currMusic->fOffset, SEEK_SET);
	SoundData* toPlay = malloc(sizeof(SoundData));
	currMusicData = sd;
	currMusic->samples = malloc(4096);
	//printf("%i\n", currMusic->samples);

	toPlay->sound = sd.sound;
	toPlay->pan = sd.pan;
	toPlay->pitch = sd.pitch;
	toPlay->volume = sd.volume;
	toPlay->loop = true;
	toPlay->loopStart = 0;
	toPlay->loopEnd = 4096 / sd.sound->bytesPerSample;

	MusicCallback(4096 / sd.sound->bytesPerSample, (char*)sd.sound->samples);

	prevTimerPos = 0;
	prevTimerRemainder = 0;
	prevMusicRemainder = 0;

	int tmp = currMusic->dataSize;
	currMusic->dataSize = 4096;

	musicId = PlaySound(toPlay);
	free(toPlay);
	currMusic->dataSize = tmp;
	if (musicId == -1) {
		StopMusic();
		playingMusic = false;
		return;
	}
	// start a timer to trigger every 1024 samples
	timerStart(0, ClockDivider_1024, 65536 - 0x2000000 / (0x1000000 / (0x1000000 / sd.sound->sampleRate)), MusicTimerCallback);
	playingMusic = true;
}

void InitializeAudio() {
	soundEnable();
}

void StopMusic() {
	if (currMusic != NULL) {
		// this doesn't get freed by the shared code...
		free(currMusic->samples);
		DestroySoundEffect(currMusic);
	}
	timerStop(0);
	timerStop(1);
}

void UninitializeAudio() {
	soundDisable();
}


int PlaySound(SoundData* sound) {
	FIFOData.data = sound;
	FIFOData.done = false;
	FIFOData.id = -1;
	fifoSendAddress(FIFO_USER_01, &FIFOData);
	while (!fifoCheckValue32(FIFO_USER_01)) {
		
	}
	int soundValue = fifoGetValue32(FIFO_USER_01);
	if (soundValue == -1) {
		return -1;
	}
	sfxIds[soundValue] = currSoundId;
	return currSoundId++;
}
#endif

SoundEffect* LoadWav(char* input) {
	char* fileName = DirToNative(input);
	FILE* f = fopen(fileName, "rb");
	free(fileName);
	if (f == NULL) {
		return NULL;
	}
	WavHeader wavHeader;
	fread(&wavHeader, sizeof(wavHeader), 1, f);
	SoundEffect* audioInfo = (SoundEffect*)malloc(sizeof(SoundEffect));
	audioInfo->sampleRate = wavHeader.sampleRate;
	audioInfo->dataSize = wavHeader.dataSize;
	audioInfo->stereo = wavHeader.channelCount == 2;
	audioInfo->bytesPerSample = wavHeader.bitsPerSample / 8;
	audioInfo->samples = (unsigned char*)malloc(wavHeader.dataSize);
	fread(audioInfo->samples, wavHeader.dataSize, 1, f);
	fclose(f);
	audioInfo->streamed = false;
	audioInfo->ogg = false;
	return audioInfo;
}

SoundEffect* LoadWavStreamed(char* input) {
	char* fileName = DirToNative(input);
	FILE* f = fopen(fileName, "rb");
	free(fileName);
	if (f == NULL) {
		return NULL;
	}
	WavHeader wavHeader;
	fread(&wavHeader, sizeof(wavHeader), 1, f);
	SoundEffect* audioInfo = (SoundEffect*)malloc(sizeof(SoundEffect));
	audioInfo->sampleRate = wavHeader.sampleRate;
	audioInfo->dataSize = wavHeader.dataSize;
	audioInfo->stereo = wavHeader.channelCount == 2;
	audioInfo->bytesPerSample = wavHeader.bitsPerSample / 8;
	audioInfo->streamFile = f;
	audioInfo->fOffset = sizeof(wavHeader);
	audioInfo->streamed = true;
	audioInfo->ogg = false;
	return audioInfo;
}

void DestroySoundEffect(SoundEffect* sound) {
	if (!sound->streamed) {
		free(sound->samples);
	}
	free(sound);
}